#
# Copyright 2025 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from .qt import QtCore, QtGui, QtWidgets

class SearchBarWidget(QtWidgets.QWidget):
    textChanged = QtCore.Signal(str)

    def __init__(self, placeholder="Search...", parent=None):
        super().__init__(parent)

        layout = QtWidgets.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.edit = QtWidgets.QLineEdit()
        self.edit.setPlaceholderText(placeholder)
        self.edit.setClearButtonEnabled(True)

        layout.addWidget(self.edit)
        self.edit.textChanged.connect(self.textChanged.emit)

    def text(self):
        return self.edit.text()

    def setText(self, text):
        self.edit.setText(text)

# Populates all validator info: name, isSuite, doc from all validator metadata
def _PopulateValidatorAndSuitesByPlugin():
    from pxr import UsdValidation
    validationRegistry = UsdValidation.ValidationRegistry()
    validatorsByPlugin = {}
    for metadata in validationRegistry.GetAllValidatorMetadata():
        # If validator doesn't belong to a specific plugin, we group these under 
        # "Uncategorized" plugin group
        plugName = metadata.plugin.name if metadata.plugin else "Uncategorized"
        validatorsByPlugin.setdefault(plugName, [])
        # Validator name implicitly has plugin name as prefix, for ui purposes
        # we will strip that out, but for uncategorized validators we want to
        # keep the full names.
        validatorName = (metadata.name.split(":")[1] 
            if plugName != "Uncategorized" else metadata.name)
        validatorsByPlugin[plugName].append(
            (validatorName, metadata.isSuite, metadata.doc))
    return validatorsByPlugin

class WordWrapDelegate(QtWidgets.QStyledItemDelegate):
    def initStyleOption(self, option, index):
        super().initStyleOption(option, index)
        option.wordWrap = True

class ValidatorPluginsWidget(QtWidgets.QWidget):
    # pluginName, validatorName, description, enabled
    validatorToggled = QtCore.Signal(str, str, str, bool) 
    def __init__(self, parent=None):
        super().__init__(parent)

        validatorsByPlugin = _PopulateValidatorAndSuitesByPlugin()

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(
            QtWidgets.QLabel(
                "Available validators/suites (grouped by plugins):"))

        self.search = SearchBarWidget("Search by plugins, validators/suites names or docs...")
        self.search.textChanged.connect(self._filterPluginsAndValidators)
        layout.addWidget(self.search)

        self.tree = QtWidgets.QTreeWidget()
        self.tree.setHeaderLabels(["Validator/Suite", "Description"])
        self.tree.setColumnCount(2)
        self.tree.setHeaderHidden(False)
        self.tree.setRootIsDecorated(True)
        self.tree.setAlternatingRowColors(True)
        self.tree.setItemDelegate(WordWrapDelegate(self.tree))
        self.tree.header().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        self.tree.header().setStretchLastSection(False)
        self.tree.itemChanged.connect(self._onItemChanged)

        allSelectorItem = self.makeCheckableItem(
            self.tree, "Select/Deselect All", role="all")
        allSelectorItem.setText(1, "Select or deselect all validators/suites.")
        allSelectorItem.setExpanded(True)
        for pluginName in validatorsByPlugin:
            pluginItem = self.makeCheckableItem(
                allSelectorItem, pluginName, role="plugin")
            if pluginName == "Uncategorized":
                pluginItem.setText(
                    1, "Validators that don't belong to a specific plugin.")
            else:
                pluginItem.setText(
                    1, f"Validators registered by {pluginName} plugin.")
            for validatorName, isSuite, doc in validatorsByPlugin[pluginName]:
                childItem = self.makeCheckableItem(pluginItem, validatorName)
                if isSuite:
                    childItem.setData(0, QtCore.Qt.ItemDataRole.UserRole, "suite")
                    doc = "(Suite) " + (doc if doc else "")
                childItem.setText(1, doc if doc else "")
                childItem.setToolTip(1, doc if doc else "")

        self.tree.sortItems(0, QtCore.Qt.SortOrder.AscendingOrder)
        self._moveUncategorizedToBottom()

        from pxr import Tf, UsdValidation
        self._noticeKeys = [
            Tf.Notice.RegisterGlobally(
                UsdValidation.Notice.DidRegisterValidator,
                self._onValidatorRegistered),
            Tf.Notice.RegisterGlobally(
                UsdValidation.Notice.DidRegisterValidatorSuite,
                self._onValidatorRegistered)
        ]

        layout.addWidget(self.tree)

    def _findPluginItem(self, pluginName):
        allSelectorItem = self.tree.invisibleRootItem().child(0)
        for i in range(allSelectorItem.childCount()):
            child = allSelectorItem.child(i)
            if child.text(0) == pluginName:
                return child
        return None

    def _findOrCreatePluginItem(self, pluginName):
        foundPluginItem = self._findPluginItem(pluginName)
        if foundPluginItem:
            return foundPluginItem
        allSelectorItem = self.tree.invisibleRootItem().child(0)
        newItem = self.makeCheckableItem(
            allSelectorItem, pluginName, role="plugin")
        if pluginName == "Uncategorized":
            newItem.setText(
                1, "Validators that don't belong to a specific plugin.")
        else:
            newItem.setText(
                1, f"Validators registered by {pluginName} plugin.")
        return newItem

    def _onValidatorRegistered(self, notice, sender):
        # When a new validator or suite is registered, we need to add it to the
        # tree. We need to insert the validator queried from the notice object
        # at appropriate position. Since all such validators are non-plugin
        # validators, we will add them under "Uncategorized" plugin group.
        from pxr import UsdValidation
        if isinstance(notice, UsdValidation.Notice.DidRegisterValidator):
            isSuite = False
            metadata = notice.GetValidator().GetMetadata()
        else:
            isSuite = True
            metadata = notice.GetValidatorSuite().GetMetadata()
        groupItem = self._findOrCreatePluginItem("Uncategorized")
        childItem = self.makeCheckableItem(groupItem, metadata.name)
        doc = metadata.doc if metadata.doc else ""
        if isSuite:
            childItem.setData(0, QtCore.Qt.ItemDataRole.UserRole, "suite")
            doc = "(Suite) " + doc
        childItem.setText(1, doc if doc else "")
        childItem.setToolTip(1, doc if doc else "")
        self.tree.sortItems(0, QtCore.Qt.SortOrder.AscendingOrder)
        self._moveUncategorizedToBottom()

    def _moveUncategorizedToBottom(self):
        allSelectorItem = self.tree.invisibleRootItem().child(0)
        for i in range(allSelectorItem.childCount()):
            if allSelectorItem.child(i).text(0) == "Uncategorized":
                allSelectorItem.addChild(allSelectorItem.takeChild(i))
                break

    def makeCheckableItem(self, parent, text, role=None):
        item = QtWidgets.QTreeWidgetItem(parent, [text])
        item.setFlags(item.flags() | QtCore.Qt.ItemFlag.ItemIsUserCheckable)
        item.setCheckState(0, QtCore.Qt.CheckState.Unchecked)
        if role == "plugin":
            item.setData(0, QtCore.Qt.ItemDataRole.UserRole, "plugin")
        elif role == "all":
            item.setData(0, QtCore.Qt.ItemDataRole.UserRole, "all")
        return item

    def _filterPluginsAndValidators(self, text):
        root = self.tree.invisibleRootItem()
        allSelectorItem = root.child(0)
        for i in range(allSelectorItem.childCount()):
            pluginItem = allSelectorItem.child(i)
            anyChildVisible = False
            if not text:
                pluginItem.setHidden(False)
                pluginItem.setExpanded(False)
                continue
            for j in range(pluginItem.childCount()):
                validatorItem = pluginItem.child(j)
                # match both name and docs
                validatorMatches = text.lower() in validatorItem.text(0).lower() or \
                                   text.lower() in validatorItem.text(1).lower()
                validatorItem.setHidden(not validatorMatches)
                validatorItem.setExpanded(validatorMatches)
                anyChildVisible = anyChildVisible or validatorMatches
            pluginItemMatches = anyChildVisible or \
                text.lower() in pluginItem.text(0).lower()
            pluginItem.setHidden(not pluginItemMatches)
            pluginItem.setExpanded(pluginItemMatches)

    def _onItemChanged(self, item, column):
        # Special toggling for everything
        if item.data(0, QtCore.Qt.ItemDataRole.UserRole) == "all":
            # Everything was changed
            state = item.checkState(0)
            for i in range(item.childCount()):
                pluginItem = item.child(i)
                pluginItem.setCheckState(0, state)
                for j in range(pluginItem.childCount()):
                    pluginItem.child(j).setCheckState(0, state)
            return
        
        # Special toggling for a specific plugin
        if item.data(0, QtCore.Qt.ItemDataRole.UserRole) == "plugin":
            state = item.checkState(0)
            for i in range(item.childCount()):
                item.child(i).setCheckState(0, state)
            return

        # toggling for a specific validator/suite
        parent = item.parent()
        if not parent:
            return
        pluginName = parent.text(0)
        validatorName = item.text(0)
        enabled = item.checkState(0) == QtCore.Qt.CheckState.Checked
        self.validatorToggled.emit(
            pluginName, validatorName, item.text(1), enabled)

class SelectedValidatorsWidget(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(QtWidgets.QLabel("Selected validators/suites:"))

        self.search = SearchBarWidget("Search by validators/suites names or docs...")
        self.search.textChanged.connect(self._filterSelectedValidators)
        layout.addWidget(self.search)

        self.tree = QtWidgets.QTreeWidget()
        self.tree.setHeaderHidden(True)
        self.tree.setRootIsDecorated(True)
        self.tree.setAlternatingRowColors(True)
        self.tree.setItemDelegate(WordWrapDelegate(self.tree))
        layout.addWidget(self.tree)

    def addValidator(self, pluginName, validatorName, description):
        pluginItem = self._findOrCreatePluginItem(pluginName)
        if not self._findChild(pluginItem, validatorName):
            item = QtWidgets.QTreeWidgetItem(pluginItem, [validatorName])
            item.setToolTip(0, description if description else "")
        self.tree.expandAll()
        self.tree.sortItems(0, QtCore.Qt.SortOrder.AscendingOrder)
        self._filterSelectedValidators(self.search.text())

    def removeValidator(self, pluginName, validatorName):
        pluginItem = self._findPluginItem(pluginName)
        if not pluginItem:
            return
        childItem = self._findChild(pluginItem, validatorName)
        if childItem:
            pluginItem.removeChild(childItem)
        if pluginItem.childCount() == 0:
            index = self.tree.indexOfTopLevelItem(pluginItem)
            self.tree.takeTopLevelItem(index)
        self.tree.sortItems(0, QtCore.Qt.SortOrder.AscendingOrder)
        self._filterSelectedValidators(self.search.text())

    def getSelectedValidators(self):
        # Returns a list of validator names in the form "PluginName:ValidatorName"
        selectedValidators = []
        for i in range(self.tree.topLevelItemCount()):
            pluginItem = self.tree.topLevelItem(i)
            pluginName = pluginItem.text(0)
            for j in range(pluginItem.childCount()):
                validatorName = pluginItem.child(j).text(0)
                if pluginName == "Uncategorized":
                    # For uncategorized validators, we keep the full name which
                    # already has plugin prefix if any.
                    selectedValidators.append(validatorName)
                else:
                    selectedValidators.append(f"{pluginName}:{validatorName}")
        return selectedValidators

    def _filterSelectedValidators(self, text):
        root = self.tree.invisibleRootItem()
        for i in range(root.childCount()):
            pluginItem = root.child(i)
            anyChildVisible = False
            if not text:
                pluginItem.setHidden(False)
                continue
            for j in range(pluginItem.childCount()):
                validatorItem = pluginItem.child(j)
                # match both name and docs
                validatorMatches = text.lower() in validatorItem.text(0).lower() or \
                                   text.lower() in validatorItem.toolTip(0).lower()
                validatorItem.setHidden(not validatorMatches)
                # expand the item if it matches
                validatorItem.setExpanded(validatorMatches)
                anyChildVisible = anyChildVisible or validatorMatches
            pluginItemMatches = anyChildVisible or \
                text.lower() in pluginItem.text(0).lower()
            pluginItem.setHidden(not pluginItemMatches)

    def _findPluginItem(self, pluginName):
        items = self.tree.findItems(pluginName, QtCore.Qt.MatchFlag.MatchExactly, 0)
        return items[0] if items else None

    def _findOrCreatePluginItem(self, pluginName):
        return self._findPluginItem(pluginName) or \
            QtWidgets.QTreeWidgetItem(self.tree, [pluginName])

    def _findChild(self, parentItem, childName):
        for i in range(parentItem.childCount()):
            if parentItem.child(i).text(0) == childName:
                return parentItem.child(i)
        return None

class TimeRangeOptions:
    def __init__(self):
        self.mode = 0
        self.startTime = 0.0
        self.endTime = 1.0
        self.includeDefaultTime = False

class OptionsWidget(QtWidgets.QWidget):
    # primsSetFromSelection, includeDescendantPrims, timeRangeOptions
    runValidation = QtCore.Signal(bool, bool, TimeRangeOptions)
    def __init__(self, parent=None):
        super().__init__(parent)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(QtWidgets.QLabel("Validation Context Options:"))

        self.primsSetFromSelection = QtWidgets.QCheckBox(
            "Validate selected prim(s)")
        self.primsSetFromSelection.setToolTip(
            "If checked, validates based on currently selected prim(s).\n"
            "If unchecked, validates the entire stage.")
        layout.addWidget(self.primsSetFromSelection)
        self.primsSetFromSelection.stateChanged.connect(
            lambda state: self.includeDescendantPrims.setEnabled(
                QtCore.Qt.CheckState(state) == QtCore.Qt.CheckState.Checked))

        self.includeDescendantPrims = QtWidgets.QCheckBox(
            "Include descendant prims")
        self.includeDescendantPrims.setToolTip(
            "If checked, includes all descendant prims of the selected prim(s).\n"
            "If unchecked, only validates the explicitly selected prim(s).")
        self.includeDescendantPrims.setEnabled(False)
        layout.addWidget(self.includeDescendantPrims)

        timeRangeGroup = QtWidgets.QGroupBox("Time Range Options")
        timeRangeLayout = QtWidgets.QVBoxLayout(timeRangeGroup)
        self.timeRangeCombo = QtWidgets.QComboBox()
        self.timeRangeCombo.addItems([
            "Current Frame",
            "Playback Range",
            "Specific Time Range"])
        timeRangeLayout.addWidget(self.timeRangeCombo)
        vLayout = QtWidgets.QVBoxLayout()
        vLayout.addWidget(QtWidgets.QLabel("Start Time:"))
        self.startTimeEdit = QtWidgets.QDoubleSpinBox()
        self.startTimeEdit.setRange(-1e6, 1e6)
        self.startTimeEdit.setValue(0.0)
        self.startTimeEdit.setEnabled(False)
        vLayout.addWidget(self.startTimeEdit)
        vLayout.addWidget(QtWidgets.QLabel("End Time:"))
        self.endTimeEdit = QtWidgets.QDoubleSpinBox()
        self.endTimeEdit.setRange(-1e6, 1e6)
        self.endTimeEdit.setValue(1.0)
        self.endTimeEdit.setEnabled(False)
        vLayout.addWidget(self.endTimeEdit)
        timeRangeLayout.addLayout(vLayout)
        self.includeDefaultTime = QtWidgets.QCheckBox("Include Default Time")
        timeRangeLayout.addWidget(self.includeDefaultTime)
        layout.addWidget(timeRangeGroup)

        def _onTimeRangeChanged(index):
            isSpecific = \
                self.timeRangeCombo.currentText() == "Specific Time Range"
            self.startTimeEdit.setEnabled(isSpecific)
            self.endTimeEdit.setEnabled(isSpecific)

        self.timeRangeCombo.currentIndexChanged.connect(_onTimeRangeChanged)

         # Add stretch to push the button to the bottom
        layout.addStretch()

        self.runButton = QtWidgets.QPushButton("Run Validation")
        self.runButton.setEnabled(False)
        layout.addWidget(self.runButton)

        self.runButton.clicked.connect(self._onRunClicked)

    def _getTimeRangeOptions(self):
        timeRangeOptions = TimeRangeOptions()
        if self.timeRangeCombo.currentText() == "Current Frame":
            timeRangeOptions.mode = 0
        elif self.timeRangeCombo.currentText() == "Playback Range":
            timeRangeOptions.mode = 1
        else:
            timeRangeOptions.mode = 2
        if timeRangeOptions.mode == 2:
            timeRangeOptions.startTime = self.startTimeEdit.value()
            timeRangeOptions.endTime = self.endTimeEdit.value()
        timeRangeOptions.includeDefaultTime = self.includeDefaultTime.isChecked()
        return timeRangeOptions


    def _onRunClicked(self):
        self.runValidation.emit(
            self.primsSetFromSelection.isChecked(), 
            self.includeDescendantPrims.isChecked(),
            self._getTimeRangeOptions())

class ErrorTypeDelegate(QtWidgets.QStyledItemDelegate):
    def paint(self, painter, option, index):
        if index.column() == 4:
            error_type = index.data()
            if error_type == "Error":
                painter.fillRect(option.rect, QtGui.QColor(220, 53, 69))
            elif error_type == "Warn":
                painter.fillRect(option.rect, QtGui.QColor(218, 165, 32))
            elif error_type == "Info":
                painter.fillRect(option.rect, QtGui.QColor(23, 162, 184))
            else:
                painter.fillRect(option.rect, QtGui.QColor(108, 117, 125))
            optionsFont = option.font
            optionsFont.setBold(True)
            option.font = optionsFont
        # Call base paint to draw text/etc.
        super().paint(painter, option, index)

class ResultWidget(QtWidgets.QWidget):
    def __init__(self, appController, parent=None):
        self._appController = appController
        super().__init__(parent)

        layout = QtWidgets.QVBoxLayout(self)       
        resultsGroupBox = QtWidgets.QGroupBox("Validation Results")
        groupBoxLayout = QtWidgets.QVBoxLayout(resultsGroupBox)
        self.search = SearchBarWidget("Search validation results by any column...")
        self.search.textChanged.connect(self._filterResultRows)
        self.search.setVisible(False)
        groupBoxLayout.addWidget(self.search)

        self.noResultLabel = QtWidgets.QLabel("No validation results yet.")
        groupBoxLayout.addWidget(self.noResultLabel)

        self.resultTable = QtWidgets.QTableWidget()
        self.resultTable.setSortingEnabled(True)
        self.resultTable.setColumnCount(5)
        self.resultTable.setHorizontalHeaderLabels(
            ["Validator Name", "Error Name", "Error Sites", "Message", "Type"])
        self.resultTable.horizontalHeader().setStretchLastSection(False)
        self.resultTable.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.Stretch)
        self.resultTable.setVisible(False)
        self.resultTable.selectionModel().selectionChanged.connect(
            self._onRowsSelected)
        self.resultTable.cellDoubleClicked.connect(self._onCellDoubleClicked)
        groupBoxLayout.addWidget(self.resultTable)
        layout.addWidget(resultsGroupBox)
        self.results = []
        self.currentSelectedErrors = []

    def populateResultsAndUpdateWidget(self, validationErrors):
        self.results.clear()
        self.resultTable.clearContents()
        if not validationErrors:
            self.search.setVisible(False)
            self.noResultLabel.setText("No validation errors found.")
            self.noResultLabel.setVisible(True)
            self.resultTable.setVisible(False)
            return

        self.search.setVisible(True)
        self.noResultLabel.setVisible(False)
        self.resultTable.setRowCount(len(validationErrors))
        self.resultTable.setVisible(True)

        for row, error in enumerate(validationErrors):
            self.results.append(error)

            nameItem = QtWidgets.QTableWidgetItem(
                error.GetValidator().GetMetadata().name)
            nameItem.setToolTip(error.GetValidator().GetMetadata().name)
            nameItem.setFlags(nameItem.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)

            errorNameItem = QtWidgets.QTableWidgetItem(error.GetName())
            errorNameItem.setToolTip(error.GetName())
            errorNameItem.setFlags(errorNameItem.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)

            errorSites = error.GetSites()
            errorSitePaths = []
            for site in errorSites:
                if site.IsPrim():
                    errorSitePaths.append(
                        site.GetPrim().GetPath().pathString)
                elif site.IsProperty():
                    errorSitePaths.append(
                        site.GetProperty().GetPath().pathString)
                else:
                    layer = site.GetLayer()
                    if layer:
                        errorSitePaths.append(layer.identifier)
                    else:
                        errorSitePaths.append(
                            error.GetStage().GetRootLayer().identifier)
            errorSiteItem = QtWidgets.QTableWidgetItem("\n".join(errorSitePaths))
            errorSiteItem.setFlags(errorSiteItem.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)

            errorMsgItem = QtWidgets.QTableWidgetItem(error.GetMessage())
            errorMsgItem.setToolTip(error.GetMessage())
            errorMsgItem.setFlags(errorMsgItem.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)

            typeItem = QtWidgets.QTableWidgetItem(error.GetType().name)
            typeItem.setFlags(typeItem.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)

            self.resultTable.setItem(row, 0, nameItem)
            self.resultTable.setItem(row, 1, errorNameItem)
            self.resultTable.setItem(row, 2, errorSiteItem)
            if len(errorSitePaths) > 1:
                self.resultTable.resizeRowToContents(row)
            self.resultTable.setItem(row, 3, errorMsgItem)
            self.resultTable.setItem(row, 4, typeItem)
            self.resultTable.setItemDelegateForColumn(4, ErrorTypeDelegate(self.resultTable))

        header = self.resultTable.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.Interactive)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.Interactive)
        header.setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeMode.Interactive)
        header.setSectionResizeMode(3, QtWidgets.QHeaderView.ResizeMode.Interactive)
        header.setSectionResizeMode(4, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        self._filterResultRows(self.search.text())

    def _filterResultRows(self, text):
        for row in range(self.resultTable.rowCount()):
            rowVisible = False
            for col in range(self.resultTable.columnCount()):
                item = self.resultTable.item(row, col)
                if item and text.lower() in item.text().lower():
                    rowVisible = True
                    break
            self.resultTable.setRowHidden(row, not rowVisible)

    def _onRowsSelected(self, selected, deselected):
        self.currentSelectedErrors.clear()
        for index in self.resultTable.selectionModel().selectedRows():
            row = index.row()
            if row < 0 or row >= len(self.results):
                continue
            error = self.results[row]
            self.currentSelectedErrors.append(error)

    def _onCellDoubleClicked(self, row, column):
        if row < 0 or row >= len(self.results):
            return
        error = self.results[row]
        errorSites = error.GetSites()
        if not errorSites:
            return

        self._appController._dataModel.selection.clear()

        primPaths = []
        propPaths = []
        for site in errorSites:
            if site.IsPrim():
                primPaths.append(site.GetPrim().GetPath())
            elif site.IsProperty():
                propPaths.append(site.GetProperty().GetPath())
        
        with self._appController._dataModel.selection.batchPrimChanges:
            self._appController._dataModel.selection.clearPrims()
            for primPath in primPaths:
                self._appController._dataModel.selection.addPrimPath(
                    primPath)

        if not propPaths:
            self._appController._ui.primView.FrameSelection()
            return

        # We can only select properties if these belong to the same prim (and
        # if sites have prim then there is only 1 prim)
        if primPaths and len(primPaths) > 1:
            # too many different prims
            return
        primPath = primPaths[0] if primPaths else propPaths[0].GetPrimPath()
        for prop in propPaths:
            if prop.GetPrimPath() != primPath:
                # properties from different prims
                return

        if propPaths[0].GetPrimPath() not in primPaths:
            with self._appController._dataModel.selection.batchPrimChanges:
                self._appController._dataModel.selection.clearPrims()
                self._appController._dataModel.selection.addPrimPath(
                    propPaths[0].GetPrimPath())

        with self._appController._dataModel.selection.batchPropChanges:
            self._appController._dataModel.selection.clearProps()
            for propPath in propPaths:
                if propPath.IsPropertyPath():
                    self._appController._dataModel.selection.addPropPath(
                        propPath)

        # Default selection data model behavior is not to frame the selected
        # prim but highlight the already expanded ancestor, but here we want
        # to frame the selected prim.
        self._appController._ui.primView.FrameSelection()


class ValidationWidget(QtWidgets.QWidget):
    def __init__(self, appController):
        super().__init__()
        self._appController = appController
        self.setWindowTitle("USD Validation")
        self.resize(1200, 800)

        # Main layout
        layout = QtWidgets.QVBoxLayout(self)

        # Vertical splitter to divide selectors/options and results
        verticalSplitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        layout.addWidget(verticalSplitter)

        # Horizontal splitter for plugins and selected/options
        topSplitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)

        self.plugins = ValidatorPluginsWidget()
        topSplitter.addWidget(self.plugins)

        # Right side container for selected validators and options, called
        # Validation Content
        rightContainer = QtWidgets.QGroupBox("Validation Context")

        rightSplitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal, rightContainer)
        
        self.selected = SelectedValidatorsWidget(rightContainer)
        rightSplitter.addWidget(self.selected)

        self.options = OptionsWidget(rightContainer)
        rightSplitter.addWidget(self.options)

        rightLayout = QtWidgets.QHBoxLayout(rightContainer)
        rightLayout.addWidget(rightSplitter)

        rightLayout.addStretch()

        topSplitter.addWidget(rightContainer)

        topSplitter.setSizes([self.width() * 0.5, self.width() * 0.5])
        topSplitter.setStretchFactor(0, 1)
        topSplitter.setStretchFactor(1, 1)
        rightLayout.setStretch(0, 1)

        verticalSplitter.addWidget(topSplitter)

        self.validationErrors = []
        self.resultWidget = ResultWidget(self._appController)
        verticalSplitter.addWidget(self.resultWidget)

        # Connect validtorToggled signal to update selected validators
        self.plugins.validatorToggled.connect(self.onValidatorToggled)

        # Connect runValidation signal to handle running validation
        self.options.runValidation.connect(self.onRunValidation)

    def updateRunButtonState(self):
        if self.selected.getSelectedValidators() and \
                not self._appController._dataModel.playing:
            self.options.runButton.setEnabled(True)
        else:
            self.options.runButton.setEnabled(False)

    def onValidatorToggled(
            self, pluginName, validatorName, description, enabled):
        if enabled:
            self.selected.addValidator(pluginName, validatorName, description)
        else:
            self.selected.removeValidator(pluginName, validatorName)
        self.updateRunButtonState()

    def onRunValidation(
            self, primsSetFromSelection, includeDescendantPrims, 
            timeRangeOptions):
        from pxr import Gf, Sdf, Usd, UsdValidation
        validatorsMetadata = \
            [UsdValidation.ValidationRegistry().GetValidatorMetadata(name) 
             for name in self.selected.getSelectedValidators()]
        validationContext = UsdValidation.ValidationContext(
            metadata=validatorsMetadata)
        stage = self._appController._dataModel.stage
        if not (validationContext or stage):
            return

        primsToValidate = []
        if primsSetFromSelection:
            primPaths = self._appController._dataModel.selection.getPrimPaths()
            # If absolute root is in the explicitly picked paths, we should do
            # the whole stage anyhow, ignore the rest.
            if not Sdf.Path.absoluteRootPath in primPaths:
                for path in primPaths:
                    primsToValidate.append(stage.GetPrimAtPath(path))

        validationTimeRange = None
        isDefaultTimeIncluded = timeRangeOptions.includeDefaultTime
        if timeRangeOptions.mode == 0:
            validationTimeRange = UsdValidation.ValidationTimeRange(
                Gf.Interval(self._appController._dataModel.currentFrame.GetValue()),
                isDefaultTimeIncluded)
        elif timeRangeOptions.mode == 1:
            validationTimeRange = UsdValidation.ValidationTimeRange(
                Gf.Interval(self._appController._dataModel.frameRangeBegin, 
                            self._appController._dataModel.frameRangeEnd),
                isDefaultTimeIncluded)
        else:
            validationTimeRange = UsdValidation.ValidationTimeRange(
                Gf.Interval(timeRangeOptions.startTime, 
                            timeRangeOptions.endTime),
                isDefaultTimeIncluded)

        if primsToValidate:
            if includeDescendantPrims:
                for prim in primsToValidate:
                    self.validationErrors.extend(
                        validationContext.Validate(Usd.PrimRange(prim)))
            else:
                self.validationErrors = validationContext.Validate(
                    primsToValidate)
        else:
            self.validationErrors = validationContext.Validate(stage)

        self.resultWidget.populateResultsAndUpdateWidget(self.validationErrors)



#
# Copyright 2017 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

"""A module for creating groups of constants. This is similar to the enum
module, but enum is not available in Python 2's standard library.
"""
import sys
import types

class _MetaConstantsGroup(type):
    """A meta-class which handles the creation and behavior of ConstantsGroups.
    """

    def __new__(metacls, cls, bases, classdict):
        """Discover constants and create a new ConstantsGroup class."""

        # If this is just the base ConstantsGroup class, simply create it and do
        # not search for constants.
        if cls == "ConstantsGroup":
            return super(_MetaConstantsGroup, metacls).__new__(metacls, cls, bases, classdict)

        # Search through the class-level properties and convert them into
        # constants.
        allConstants = list()
        for key, value in classdict.items():
            if (key.startswith("_") or isinstance(value, classmethod) or
                    isinstance(value, staticmethod)):
                # Ignore variables which start with an underscore, and
                # static/class methods.
                pass
            else:
                # Found a new constant.
                allConstants.append(value)

                # If the constant is a function/lambda, ensure that it is not
                # converted into a method.
                if isinstance(value, types.FunctionType):
                    classdict[key] = staticmethod(value)

        # All constants discovered, now create the `_all` property.
        classdict["_all"] = tuple(allConstants)

        # Finally, create the new ConstantsGroup class.
        return super(_MetaConstantsGroup, metacls).__new__(metacls, cls, bases, classdict)

    def __setattr__(cls, name, value):
        """Prevent modification of properties after a group is created."""
        raise AttributeError("Constant groups cannot be modified.")

    def __delattr__(cls, name):
        """Prevent deletion of properties after a group is created."""
        raise AttributeError("Constant groups cannot be modified.")

    def __len__(self):
        """Get the number of constants in the group."""
        return len(self._all)

    def __contains__(self, value):
        """Check if a constant exists in the group."""
        return (value in self._all)

    def __iter__(self):
        """Iterate over each constant in the group."""
        return iter(self._all)

# We want to define a ConstantsGroup class that uses _MetaConstantsGroup
# as its metaclass.
class ConstantsGroup(object, metaclass=_MetaConstantsGroup):
    """The base constant group class, intended to be inherited by actual groups
    of constants.
    """
    def __new__(cls, *args, **kwargs):
        raise TypeError("ConstantsGroup objects cannot be created.")



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from pxr import Tf
Tf.PreparePythonModule()
del Tf

# Utilities for unit testing

def _TestPrimIndex(primIndex, expected):
    '''Generator for verifying the expected structure and
    values throughout the given prim index.

    The "expected" parameter is a list that mirrors the tree
    structure of the prim index:

    expected  : [nodeEntry]
    nodeEntry : (tuple of arbitrary data), [nodeEntry, ...]

    The first item in a nodeEntry is a tuple of arbitrary data
    associated with a node in the prim index. The second item is
    a list of nodeEntries corresponding to the children of that
    node.

    This generator will visit each node in the prim index and
    yield the node and the associated data tuple.

    For example:

    expected = [
        (Pcp.ArcTypeRoot, "/Root"), [
            (Pcp.ArcTypeReference, "/Ref1"), [],
            (Pcp.ArcTypeReference, "/Ref2"), []
        ]
    ]
          
    for node, entry in Pcp._TestPrimIndex(primIndex, expected):
        assert node.arcType == entry.arcType
        assert node.path == entry.path

    '''
    def _recurse(node, expected):
        try:
            yield node, expected[0]
        except IndexError as e:
            raise RuntimeError(
                "No entry in expected corresponding to node {}"
                .format(node.site)) from e

        if len(expected[1]) % 2 != 0:
            raise RuntimeError(
                "Invalid children list. Each element in the list "
                "must be a tuple of node data followed by another list")

        numExpectedChildren = len(expected[1]) // 2
        if len(node.children) != numExpectedChildren:
            raise RuntimeError(
                "Expected {} children but have {} for node {}"
                .format(numExpectedChildren, len(node.children), node.site))

        for idx, n in enumerate(node.children):
            try:
                expectedSubtree = expected[1][(idx*2):(idx*2)+2]
            except IndexError as e:
                raise RuntimeError(
                    "No entry in expected corresponding to node {}"
                    .format(n.site)) from e
            yield from _recurse(n, expectedSubtree)

    yield from _recurse(primIndex.rootNode, expected)



#
# Copyright 2017 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from .qt import QtCore, QtWidgets
from .adjustDefaultMaterialUI import Ui_AdjustDefaultMaterial

class AdjustDefaultMaterial(QtWidgets.QDialog):
    """Popup widget to adjust the default material used for rendering.
    `datamodel` should be a ViewSettingsDataModel.
    """

    def __init__(self, parent, dataModel):
        QtWidgets.QDialog.__init__(self,parent)
        self._ui = Ui_AdjustDefaultMaterial()
        self._ui.setupUi(self)

        self._dataModel = dataModel
        self._ambientCache = None
        self._specularCache = None

        self._ui.ambientIntSpinBox.valueChanged['double'].connect(self._ambientChanged)

        self._ui.specularIntSpinBox.valueChanged['double'].connect(self._specularChanged)

        dataModel.signalDefaultMaterialChanged.connect(self._updateFromData)

        self._ui.resetButton.clicked[bool].connect(self._reset)

        self._ui.doneButton.clicked[bool].connect(self._done)

        self._updateFromData()

    def _updateFromData(self):
        if self._dataModel.defaultMaterialAmbient != self._ambientCache:
            self._ambientCache = self._dataModel.defaultMaterialAmbient
            self._ui.ambientIntSpinBox.setValue(self._ambientCache)

        if self._dataModel.defaultMaterialSpecular != self._specularCache:
            self._specularCache = self._dataModel.defaultMaterialSpecular
            self._ui.specularIntSpinBox.setValue(self._specularCache)


    def _ambientChanged(self, val):
        if val != self._ambientCache:
            # Must do update cache first to prevent update cycle
            self._ambientCache = val
            self._dataModel.defaultMaterialAmbient = val

    def _specularChanged(self, val):
        if val != self._specularCache:
            # Must do update cache first to prevent update cycle
            self._specularCache = val
            self._dataModel.defaultMaterialSpecular = val

    def _reset(self, unused):
        self._dataModel.resetDefaultMaterial()

    def _done(self, unused):
        self.close()

    def closeEvent(self, event):
        event.accept()
        # Since the dialog is the immediate-edit kind, we consider
        # window-close to be an accept, so our clients can know the dialog is
        # done
        self.accept()



#
# Copyright 2017 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
""" This module provides the help dialog(also known as the property legend)
    in Usdview's MainWindow. This provides a key describing the items displayed
    in the property browser.
"""

from .qt import QtWidgets

from .propertyLegendUI import Ui_PropertyLegend

from .common import UIBaseColors, UIPropertyValueSourceColors, ItalicizeLabelText, PropertyViewIcons

class PropertyLegend(QtWidgets.QWidget):
    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent)
        self._ui = Ui_PropertyLegend()
        self._ui.setupUi(self)

        # The property legend always starts off collapsed.
        self.setMaximumHeight(0)
        self._isMinimized = True
        self._iconDisplaySize = (16, 16)
        
        graphicsScene = QtWidgets.QGraphicsScene()

        self._ui.propertyLegendColorFallback.setScene(graphicsScene)
        self._ui.propertyLegendColorDefault.setScene(graphicsScene)
        self._ui.propertyLegendColorTimeSample.setScene(graphicsScene)
        self._ui.propertyLegendColorNoValue.setScene(graphicsScene)
        self._ui.propertyLegendColorValueClips.setScene(graphicsScene)
        self._ui.propertyLegendColorCustom.setScene(graphicsScene)
        self._ui.propertyLegendColorSpline.setScene(graphicsScene)

        # set color of attribute viewer legend boxes
        self._ui.propertyLegendColorFallback.setForegroundBrush(
            UIPropertyValueSourceColors.FALLBACK)
        self._ui.propertyLegendColorDefault.setForegroundBrush(
            UIPropertyValueSourceColors.DEFAULT)
        self._ui.propertyLegendColorTimeSample.setForegroundBrush(
            UIPropertyValueSourceColors.TIME_SAMPLE)
        self._ui.propertyLegendColorNoValue.setForegroundBrush(
            UIPropertyValueSourceColors.NONE)
        self._ui.propertyLegendColorValueClips.setForegroundBrush(
            UIPropertyValueSourceColors.VALUE_CLIPS)
        self._ui.propertyLegendColorCustom.setForegroundBrush(
            UIBaseColors.RED)
        self._ui.propertyLegendColorSpline.setForegroundBrush(
            UIPropertyValueSourceColors.SPLINE)

        # set color of attribute viewer text items
        legendTextUpdate = lambda t, c: (
            ('<font color=\"%s\">' % c.color().name()) + t.text() + '</font>')
        timeSampleLegend = self._ui.propertyLegendLabelTimeSample
        timeSampleLegend.setText(
            legendTextUpdate(timeSampleLegend, UIPropertyValueSourceColors.TIME_SAMPLE))

        fallbackLegend = self._ui.propertyLegendLabelFallback
        fallbackLegend.setText(
            legendTextUpdate(fallbackLegend, UIPropertyValueSourceColors.FALLBACK))

        valueClipLegend = self._ui.propertyLegendLabelValueClips
        valueClipLegend.setText(
            legendTextUpdate(valueClipLegend, UIPropertyValueSourceColors.VALUE_CLIPS))

        noValueLegend = self._ui.propertyLegendLabelNoValue
        noValueLegend.setText(
            legendTextUpdate(noValueLegend, UIPropertyValueSourceColors.NONE))

        defaultLegend = self._ui.propertyLegendLabelDefault
        defaultLegend.setText(
            legendTextUpdate(defaultLegend, UIPropertyValueSourceColors.DEFAULT))

        customLegend = self._ui.propertyLegendLabelCustom
        customLegend.setText(
            legendTextUpdate(customLegend, UIBaseColors.RED))

        splineLegend = self._ui.propertyLegendLabelSpline
        splineLegend.setText(
            legendTextUpdate(splineLegend, UIPropertyValueSourceColors.SPLINE))

        interpolatedStr = 'Interpolated'
        tsLabel = self._ui.propertyLegendLabelTimeSample
        tsLabel.setText(ItalicizeLabelText(tsLabel.text(), interpolatedStr))

        vcLabel = self._ui.propertyLegendLabelValueClips
        vcLabel.setText(ItalicizeLabelText(vcLabel.text(), interpolatedStr))

        splineLabel = self._ui.propertyLegendLabelSpline
        splineLabel.setText(
            ItalicizeLabelText(splineLabel.text(), interpolatedStr))

        # Load up and set the icons for the property legend
        self._ui.propertyLegendTargetIcon.setPixmap(
            PropertyViewIcons.TARGET().pixmap(*self._iconDisplaySize))
        self._ui.propertyLegendConnIcon.setPixmap(
            PropertyViewIcons.CONNECTION().pixmap(*self._iconDisplaySize))
        self._ui.propertyLegendAttrPlainIcon.setPixmap(
            PropertyViewIcons.ATTRIBUTE().pixmap(*self._iconDisplaySize))
        self._ui.propertyLegendRelPlainIcon.setPixmap(
            PropertyViewIcons.RELATIONSHIP().pixmap(*self._iconDisplaySize))
        self._ui.propertyLegendAttrWithConnIcon.setPixmap(
            PropertyViewIcons.ATTRIBUTE_WITH_CONNECTIONS().pixmap(*self._iconDisplaySize))
        self._ui.propertyLegendRelWithTargetIcon.setPixmap(
            PropertyViewIcons.RELATIONSHIP_WITH_TARGETS().pixmap(*self._iconDisplaySize))
        self._ui.propertyLegendCompIcon.setPixmap(
            PropertyViewIcons.COMPOSED().pixmap(*self._iconDisplaySize))

    def IsMinimized(self):
        return self._isMinimized

    def ToggleMinimized(self):
        self._isMinimized = not self._isMinimized

    def GetHeight(self):
        return self.height()

    def GetResetHeight(self):
        return self.sizeHint().height()



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

from __future__ import print_function

from pxr import Tf
Tf.PreparePythonModule()

import sys, argparse, os

from .qt import QtCore, QtGui, QtWidgets
from .common import Timer, GetIconPath
from .appController import AppController
from .settings import ConfigManager

from pxr import Sdf, UsdUtils, UsdAppUtils


class InvalidUsdviewOption(Exception):
    """Raised when an invalid Usdview option is found in
    Launcher.ValidateOptions or any methods which override it.
    """
    pass


def _AbsoluteFilePath(argStr):
    """This should be used for args that are meant to absolute file paths (as
    opposed to ones that may get resolved.  This is especially useful if you
    have an app that uses this launcher but also changes the dir.
    """
    if argStr is None:
        return None
    return os.path.abspath(argStr)


class Launcher(object):
    '''
    Base class for argument parsing, validation, and initialization for UsdView

    Subclasses can choose to override
      -- GetHelpDescription()
      -- RegisterOptions()
      -- ParseOptions()
      -- ValidateOptions()
      -- GetResolverContext()
    '''
    def __init__(self):
        pass

    def Run(self):
        '''
        The main entry point to launch a process using UsdView.
        '''

        parser = argparse.ArgumentParser(prog=sys.argv[0],
                                         description=self.GetHelpDescription())

        traceCollector = None

        with Timer('open and close usdview') as totalTimer:
            self.RegisterPositionals(parser)
            self.RegisterOptions(parser)
            arg_parse_result = self.ParseOptions(parser)
            self.ValidateOptions(arg_parse_result)

            if arg_parse_result.traceToFile:
                from pxr import Trace
                traceCollector = Trace.Collector()
            
                if arg_parse_result.tracePython:
                    traceCollector.pythonTracingEnabled = True

                traceCollector.enabled = True

            app, appController = self.LaunchPreamble(arg_parse_result)

            if arg_parse_result.viewportSize and appController._stageView:
                w, h = arg_parse_result.viewportSize
                appController._stageView.SetPhysicalWindowSize(w, h)

            if arg_parse_result.dumpFirstImage:
                appController.SaveViewerImageToFile(arg_parse_result.dumpFirstImage)

            self.LaunchProcess(arg_parse_result, app, appController)

        if traceCollector:
            traceCollector.enabled = False

        if arg_parse_result.timing and arg_parse_result.quitAfterStartup:
            totalTimer.PrintTime()

        if traceCollector:
            if arg_parse_result.traceFormat == 'trace':
                Trace.Reporter.globalReporter.Report(
                    arg_parse_result.traceToFile)
            elif arg_parse_result.traceFormat == 'chrome':
                Trace.Reporter.globalReporter.ReportChromeTracingToFile(
                    arg_parse_result.traceToFile)
            else:
                Tf.RaiseCodingError("Invalid trace format option provided: %s -"
                        "trace/chrome are the valid options" %
                        arg_parse_result.traceFormat)


    def GetHelpDescription(self):
        '''return the help description'''
        return 'View a usd file'

    def RegisterPositionals(self, parser):
        '''
        register positional arguments on the ArgParser
        '''
        parser.add_argument('usdFile', action='store',
                            nargs='?',
                            type=str,
                            help='The file to view (Optional)')

    def RegisterOptions(self, parser):
        '''
        register optional arguments on the ArgParser
        '''
        from pxr import UsdUtils

        UsdAppUtils.rendererArgs.AddCmdlineArgs(parser,
                altHelpText=("Which render backend to use (named as it "
                            "appears in the menu). 'GL' and 'Storm' currently"
                            "alias to the same renderer, Storm."))
        
        parser.add_argument('--select', action='store', default='/',
                            dest='primPath', type=str,
                            help='A prim path to initially select and frame')

        UsdAppUtils.cameraArgs.AddCmdlineArgs(parser,
            altHelpText=(
                "Which camera to set the view to on open - may be given as "
                "either just the camera's prim name (ie, just the last "
                "element in the prim path), or as a full prim path.  Note "
                "that if only the prim name is used, and more than one camera "
                "exists with the name, which is used will be effectively "
                "random (default=%(default)s)"))

        parser.add_argument('--mask', action='store',
                            dest='populationMask',
                            metavar='PRIMPATH[,PRIMPATH...]',
                            help='Limit stage population to these prims, '
                            'their descendants and ancestors.  To specify '
                            'multiple paths, either use commas with no spaces '
                            'or quote the argument and separate paths by '
                            'commas and/or spaces.')

        parser.add_argument('--clearsettings', action='store_true',
                            dest='clearSettings',
                            help='Restores usdview settings to default')

        parser.add_argument('--config', action='store',
                            type=str,
                            dest='config',
                            default=ConfigManager.defaultConfig,
                            choices=ConfigManager(
                                AppController._outputBaseDirectory()
                            ).getConfigs()[1:],
                            help='Load usdview with the state settings found '
                            'in the specified config. If not provided will '
                            'use the previously saved application state and '
                            'automatically persist state on close')

        parser.add_argument('--defaultsettings', action='store_true',
                            dest='defaultSettings',
                            help='Launch usdview with default settings')

        parser.add_argument('--norender', action='store_true',
                            dest='noRender',
                            help='Display only hierarchy browser')

        parser.add_argument('--noplugins', action='store_true',
                            dest='noPlugins',
                            help='Do not load plugins')

        parser.add_argument('--unloaded', action='store_true',
                            dest='unloaded',
                            help='Do not load payloads')

        parser.add_argument('--bboxStandin', action='store_true',
                            dest='bboxstandin',
                            help='Display unloaded prims with bounding boxes')

        parser.add_argument('--timing', action='store_true',
                            dest='timing',
                            help='Echo timing stats to console. NOTE: timings will be unreliable when the --memstats option is also in use')

        parser.add_argument('--allow-async', action='store_true',
                            dest='allowAsync',
                            help='Enable asynchronous hydra scene processing')

        parser.add_argument('--traceToFile', action='store',
                            type=_AbsoluteFilePath,
                            dest='traceToFile',
                            default=None,
                            help='Start tracing at application startup and '
                            'write --traceFormat specified format output to the '
                            'specified trace file when the application quits')

        parser.add_argument('--traceFormat', action='store',
                            type=str,
                            dest='traceFormat',
                            default='chrome',
                            choices=['chrome', 'trace'],
                            help='Output format for trace file specified by '
                            '--traceToFile. \'chrome\' files can be read in '
                            'chrome, \'trace\' files are simple text reports. '
                            '(default=%(default)s)')

        parser.add_argument('--tracePython', action='store_true',
                            dest='tracePython',
                            help='Enable python trace collection, '
                            'requires --traceToFile to be set.')

        parser.add_argument('--memstats', action='store', default='none',
                            dest='mallocTagStats', type=str,
                            choices=['none', 'stage', 'stageAndImaging'],
                            help='Use the Pxr MallocTags memory accounting system to profile USD, saving results to a tmp file, with a summary to the console.  Will have no effect if MallocTags are not supported in the USD installation.')

        parser.add_argument('--dumpFirstImage', action='store',
                            type=_AbsoluteFilePath,
                            dest='dumpFirstImage',
                            default=None,
                            help='Dumps the first image to file (as png)')

        parser.add_argument('--viewportSize', action='store',
                            type=lambda s: tuple(int(x) for x in s.split('x')),
                            dest='viewportSize',
                            default=None,
                            help='Fix the viewport to a specific size (WxH) '
                                 'for deterministic rendering, e.g. 601x458')

        parser.add_argument('--numThreads', action='store',
                            type=int, default=0,
                            help='Number of threads used for processing'
                                 '(0 is max, negative numbers imply max - N)')

        parser.add_argument('--ff', action='store',
                            dest='firstframe', type=int,
                            help='Set the first frame of the viewer')

        parser.add_argument('--lf', action='store',
                            dest='lastframe', type=int,
                            help='Set the last frame of the viewer')

        parser.add_argument('--cf', action='store',
                            dest='currentframe', type=int,
                            help='Set the current frame of the viewer')

        UsdAppUtils.complexityArgs.AddCmdlineArgs(parser,
            altHelpText=(
                'Set the initial mesh refinement complexity (%(default)s).'))

        parser.add_argument('--quitAfterStartup', action='store_true',
                            dest='quitAfterStartup',
                            help='quit immediately after start up')
                            
        parser.add_argument('--sessionLayer', default=None, type=str,
                            help= "If specified, the stage will be opened "
                            "with the 'sessionLayer' in place of the default "
                            "anonymous layer. As this changes the session "
                            "layer from anonymous to persistent, be "
                            "aware that layers saved from Export Overrides "
                            "will include the opinions in the persistent "
                            "session layer.")

        parser.add_argument('--mute', default=None, type=str,
                            dest='muteLayersRe', action='append', nargs=1,
                            help="Layer identifiers searched against this "
                                 "regular expression will be muted on the "
                                 "stage prior to, and after loading. Multiple "
                                 "expressions can be supplied using the | "
                                 "regex separator operator. Alternatively the "
                                 "argument may be used multiple times.")

        group = parser.add_argument_group(
            'Detached Layers',
            'Specify layers to be detached from their serialized data source '
            'when loaded. This may increase time to load and memory usage but '
            'will avoid issues like open file handles preventing other '
            'processes from safely overwriting a loaded layer.')

        group.add_argument(
            '--detachLayers', action='store_true', help=("Detach all layers"))

        group.add_argument(
            '--detachLayersInclude', action='store', 
            metavar='PATTERN[,PATTERN...]',
            help=("Detach layers with identifiers containing any of the "
                  "given patterns."))

        group.add_argument(
            '--detachLayersExclude', action='store',
            metavar='PATTERN[,PATTERN,...]',
            help=("Exclude layers with identifiers containing any of the "
                  "given patterns from the set of detached layers specified "
                  "by the --detachLayers or --detachLayerIncludes arguments."))

    def ParseOptions(self, parser):
        '''
        runs the parser on the arguments
        '''
        return parser.parse_args()

    def ValidateOptions(self, arg_parse_result):
        '''
        Called by Run(), after ParseOptions() is called. Validates and 
        potentially modifies the parsed arguments. Raises InvalidUsdviewOption 
        if an invalid option is found. If a derived class has overridden 
        ParseOptions(), ValidateOptions() is an opportunity to process the
        options and transmute other "core" options in response.  If 
        overridden, derived classes should likely first call the base method.
        '''

        # Split arg_parse_result.populationMask into paths.
        if arg_parse_result.populationMask:
            arg_parse_result.populationMask = (
                arg_parse_result.populationMask.replace(',', ' ').split())

        # Process detached layer arguments.
        if arg_parse_result.detachLayersInclude:
            arg_parse_result.detachLayersInclude = [
                s for s in arg_parse_result.detachLayersInclude.split(',') if s
            ]

        if arg_parse_result.detachLayersExclude:
            arg_parse_result.detachLayersExclude = [
                s for s in arg_parse_result.detachLayersExclude.split(',') if s
            ]

        # Verify that the camera path is either an absolute path, or is just
        # the name of a camera.
        if not arg_parse_result.camera:
            from pxr import Sdf
            primaryCameraName = UsdUtils.GetPrimaryCameraName()
            if primaryCameraName:
                arg_parse_result.camera = Sdf.Path(primaryCameraName)
            else:
                arg_parse_result.camera = Sdf.Path.emptyPath
        if arg_parse_result.camera:
            camPath = arg_parse_result.camera
            if camPath.isEmpty:
                raise InvalidUsdviewOption(
                    "invalid camera path - %r" % camPath)
            if not camPath.IsPrimPath():
                raise InvalidUsdviewOption(
                    "invalid camera path - must be a raw prim path with no "
                    "variant selections or properties - got: %r" % camPath)

            # If it's a multi-element path, make sure it is absolute.
            if camPath.pathElementCount > 1:
                if not camPath.IsAbsolutePath():
                    # perhaps we should error here? For now just pre-pending
                    # root, and printing warning...
                    from pxr import Sdf
                    print("WARNING: camera path %r was not absolute, prepending "
                          "%r to make it absolute" % (str(camPath),
                              str(Sdf.Path.absoluteRootPath)), file=sys.stderr)
                    arg_parse_result.camera = camPath.MakeAbsolutePath(
                        Sdf.Path.absoluteRootPath)

        if arg_parse_result.clearSettings and arg_parse_result.defaultSettings:
            raise InvalidUsdviewOption(
                "cannot supply both --clearsettings and --defaultsettings.")

    def GetResolverContext(self, usdFile):
        """
        Create and return the ArResolverContext that will be used to Open
        the Stage for the given usdFile.  Base implementation
        creates a default asset context for the usdFile asset, but derived
        classes can do more sophisticated resolver and context configuration.
        
        Will be called each time a new stage is opened.

        It is not necessary to create an ArResolverContext for every UsdStage
        one opens, as the Stage will use reasonable fallback behavior if no
        context is provided.  For usdview, configuring an asset context by
        default is reasonable, and allows clients that embed usdview to 
        achieve different behavior when needed.
        
        If usdFile path is not provided, it returns default context.
        """
        from pxr import Ar
        
        r = Ar.GetResolver()
        
        if not usdFile:
            return r.CreateDefaultContext()

        # ConfigureResolverForAsset no longer exists under Ar 2.0; this
        # is here for backwards compatibility with Ar 1.0.
        if hasattr(r, "ConfigureResolverForAsset"):
            r.ConfigureResolverForAsset(usdFile)

        return r.CreateDefaultContextForAsset(usdFile)

    def OverrideMaxSamples(self):
        # Return True if HdPrman's default max samples should be overridden,
        # False otherwise.
        return True

    def LaunchPreamble(self, arg_parse_result):
        # Initialize concurrency limit as early as possible so that it is
        # respected by subsequent imports.
        from pxr import Work
        Work.SetConcurrencyLimitArgument(arg_parse_result.numThreads)

        # XXX Override HdPrman's defaults using the env var.  In the
        # future we expect there may be more formal ways to represent
        # per-app settings for particular Hydra plugins.
        if self.OverrideMaxSamples():
            os.environ.setdefault('HD_PRMAN_MAX_SAMPLES', '1024')

        if arg_parse_result.clearSettings:
            AppController.clearSettings()

        # On Windows, set the AppUserModelID so that the taskbar groups
        # usdview as its own application rather than under "Python".
        if sys.platform == 'win32':
            import ctypes
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(
                'pxr.usd.usdview')

        # On MacOS, override the bundle name so the menu bar and Dock
        # show "usdview" rather than "Python"
        if sys.platform == 'darwin':
            import ctypes
            import ctypes.util
            objc_lib = ctypes.util.find_library('objc')
            if objc_lib:
                objc = ctypes.cdll.LoadLibrary(objc_lib)
                objc.objc_getClass.restype = ctypes.c_void_p
                objc.sel_registerName.restype = ctypes.c_void_p
                objc.objc_msgSend.restype = ctypes.c_void_p
                objc.objc_msgSend.argtypes = [
                    ctypes.c_void_p, ctypes.c_void_p]

                NSBundle = objc.objc_getClass(b'NSBundle')
                mainBundle = objc.objc_msgSend(
                    NSBundle, objc.sel_registerName(b'mainBundle'))
                info = objc.objc_msgSend(
                    mainBundle, objc.sel_registerName(b'infoDictionary'))

                objc.objc_msgSend.argtypes = [
                    ctypes.c_void_p, ctypes.c_void_p, ctypes.c_char_p]
                NSString = objc.objc_getClass(b'NSString')
                strSel = objc.sel_registerName(b'stringWithUTF8String:')
                setSel = objc.sel_registerName(b'setObject:forKey:')

                for key in (b'CFBundleName', b'CFBundleDisplayName'):
                    objc.objc_msgSend.argtypes = [
                        ctypes.c_void_p, ctypes.c_void_p,
                        ctypes.c_char_p]
                    nsKey = objc.objc_msgSend(NSString, strSel, key)
                    nsVal = objc.objc_msgSend(NSString, strSel, b'usdview')
                    objc.objc_msgSend.argtypes = [
                        ctypes.c_void_p, ctypes.c_void_p,
                        ctypes.c_void_p, ctypes.c_void_p]
                    objc.objc_msgSend(info, setSel, nsVal, nsKey)

        # Create the Qt application
        app = QtWidgets.QApplication(sys.argv)
        app.setApplicationName("usdview")
        app.setApplicationDisplayName("usdview")

        contextCreator = lambda usdFile: self.GetResolverContext(usdFile)
        appController = AppController(arg_parse_result, contextCreator)

        return (app, appController)

    def LaunchProcess(self, arg_parse_result, app, appController):
        '''
        after the arguments have been parsed, launch the UI in a forked process
        '''
        
        if arg_parse_result.quitAfterStartup:
            # Enqueue event to shutdown application. We don't use quit() because
            # it doesn't trigger the closeEvent() on the main window which is
            # used to orchestrate the shutdown.
            QtCore.QTimer.singleShot(0, app.instance().closeAllWindows)

        app.exec_()




#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from pxr import Tf

Tf.ScriptModuleLoader()._LoadModulesForLibrary('LoadsUnknown')

from . import testTfScriptModuleLoader_Unknown



#
# Copyright 2024 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
"""
Level of Detail (LOD) API schemas for USD.
"""

from pxr import Tf
Tf.PreparePythonModule()
del Tf



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
"""Vt Value Types library

This package defines classes for creating objects that behave like value
types.  It contains facilities for creating simple copy-on-write
implicitly shared types.
"""

from pxr import Tf
Tf.PreparePythonModule()
del Tf

# For ease-of-use, put each XXXArrayFromBuffer method on its corresponding array
# wrapper class, also alias it as 'FromNumpy' for compatibility.
def _CopyArrayFromBufferFuncs(moduleContents):
    funcs = dict([(key, val) for (key, val) in moduleContents.items()
                  if key.endswith('FromBuffer')])
    classes = dict([(key, val) for (key, val) in moduleContents.items()
                    if key.endswith('Array') and isinstance(val, type)])

    for funcName, func in funcs.items():
        className = funcName[:-len('FromBuffer')]
        cls = classes.get(className)
        if cls:
            setattr(cls, 'FromBuffer', staticmethod(func))
            setattr(cls, 'FromNumpy', staticmethod(func))

_CopyArrayFromBufferFuncs(dict(vars()))
del _CopyArrayFromBufferFuncs



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from .qt import QtCore, QtGui, QtWidgets
from .preferencesUI import Ui_Preferences
from .common import FixableDoubleValidator

class Preferences(QtWidgets.QDialog):
    """The dataModel provided to this VC must conform to the following
    interface:

    Editable properties:
       fontSize, int
       appIconMode, str

    Readable properties:

    Signals:
       viewSettings.signalSettingChanged() - whenever any view setting 
                                             may have changed.
    """
    def __init__(self, parent, dataModel):
        super(Preferences, self).__init__(parent)
        self._ui = Ui_Preferences()
        self._ui.setupUi(self)

        self._dataModel = dataModel
        self._dataModel.viewSettings.signalSettingChanged.connect(self._updateEditorsFromDataModel)
        self._muteUpdates = False

        # When the checkboxes change, we want to update instantly
        self._ui.buttonBox.clicked.connect(self._buttonBoxButtonClicked)

        self._updateEditorsFromDataModel()

    def _updateEditorsFromDataModel(self):
        if self._muteUpdates:
            return
        self._ui.fontSizeSpinBox.setValue(self._dataModel.viewSettings.fontSize)
        idx = self._ui.appIconComboBox.findText(
            self._dataModel.viewSettings.appIconMode)
        if idx >= 0:
            self._ui.appIconComboBox.setCurrentIndex(idx)
        self.update()


    def _apply(self):
        self._muteUpdates = True
        self._dataModel.viewSettings.fontSize = self._ui.fontSizeSpinBox.value()
        self._dataModel.viewSettings.appIconMode = self._ui.appIconComboBox.currentText()
        self._muteUpdates = False

    def _buttonBoxButtonClicked(self, button):
        role = self._ui.buttonBox.buttonRole(button)
        Roles = QtWidgets.QDialogButtonBox.ButtonRole
        if role == Roles.AcceptRole or role == Roles.ApplyRole:
            self._apply()
        if role == Roles.AcceptRole or role == Roles.RejectRole:
            self.close()
